/*
 * ExactSubtreeAutomaton.cpp
 *
 *  Created on: 7. 4. 2015
 *      Author: Jan Travnicek
 */

#include "ExactSubtreeAutomaton.h"
#include <tree/Tree.h>
#include <tree/ranked/PrefixRankedTree.h>
#include <automaton/Automaton.h>
#include <automaton/PDA/InputDrivenNPDA.h>
#include <alphabet/RankedSymbol.h>

#include <deque>

namespace arbology {

namespace exact {

automaton::Automaton ExactSubtreeAutomaton::construct ( const tree::Tree & text ) {
	return dispatch ( text.getData ( ) );
}

automaton::InputDrivenNPDA < > ExactSubtreeAutomaton::construct ( const tree::PrefixRankedTree < > & tree ) {
	alphabet::Symbol S = alphabet::Symbol ( 'S' );
	automaton::InputDrivenNPDA < > res ( label::Label ( 0 ), S );

	for ( const std::ranked_symbol < > & rankedSymbol : tree.getAlphabet ( ) ) {
		alphabet::Symbol symbol ( alphabet::RankedSymbol < > { rankedSymbol } );
		res.addInputSymbol ( symbol );
		res.setPushdownStoreOperation ( symbol, std::vector < alphabet::Symbol > ( 1, S ), std::vector < alphabet::Symbol > ( ( size_t ) rankedSymbol.getRank ( ), S ) );
	}

	int i = 1;

	for ( const std::ranked_symbol < > & rankedSymbol : tree.getContent ( ) ) {
		alphabet::Symbol symbol ( alphabet::RankedSymbol < > { rankedSymbol } );

		res.addState ( label::Label ( i ) );
		res.addTransition ( label::Label ( i - 1 ), symbol, label::Label ( i ) );
		res.addTransition ( label::Label ( 0 ), std::move ( symbol ), label::Label ( i ) );
		i++;
	}

	return res;
}

auto ExactSubtreeAutomatonPrefixRankedTree = ExactSubtreeAutomaton::RegistratorWrapper < automaton::InputDrivenNPDA < >, tree::PrefixRankedTree < > > ( ExactSubtreeAutomaton::construct );

} /* namespace exact */

} /* namespace arbology */
