/*
 * UnreachableSymbols.h
 *
 *  Created on: 22. 3. 2014
 *	  Author: Tomas Pecka
 */

#ifndef UNREACHABLE_SYMBOLS_H_
#define UNREACHABLE_SYMBOLS_H_

#include <core/multipleDispatch.hpp>

#include <grammar/Grammar.h>
#include <alphabet/Symbol.h>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include <algorithm>
#include <deque>
#include <set>

namespace grammar {

namespace properties {

/**
 * Implements algorithms from Melichar, chapter 3.3
 */
class UnreachableSymbols : public std::SingleDispatch<UnreachableSymbols, std::set<alphabet::Symbol>, const grammar::GrammarBase &> {
public:
	static std::set<alphabet::Symbol> getUnreachableSymbols( const grammar::Grammar & grammar );

	/**
	 * Implements
	 */
	template<class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static std::set<alphabet::Symbol> getUnreachableSymbols( const T & grammar );
};

template<class T, class SymbolType >
std::set<alphabet::Symbol> UnreachableSymbols::getUnreachableSymbols( const T & grammar ) {
	// 1
	std::deque<std::set<alphabet::Symbol>> Vi;
	Vi.push_back( std::set<alphabet::Symbol>( ) );
	Vi.at( 0 ).insert( grammar.getInitialSymbol( ) );

	int i = 1;

	// 2.
	while( true ) {
		Vi.push_back( Vi.at( i - 1 ) );

		for( const auto & rule : grammar.getRawRules( ) ) {
			if( Vi.at( i - 1 ).count( rule.first ) ) {
				for( const auto & rhs : rule.second ) {
					Vi.at( i ).insert( rhs.begin( ), rhs.end( ) );
				}
			}
		}


		if( Vi.at( i ) == Vi.at( i - 1 ) )
			break;

		i = i + 1;
	}

	// 3.
	return Vi.at( i );
}

} /* namespace properties */

} /* namespace grammar */

#endif /* UNREACHABLE_SYMBOLS_H_ */
