/*
 * ToAutomatonThompson.cpp
 *
 *  Created on: 11. 1. 2014
 *	  Author: Tomas Pecka
 */
#include "ToAutomatonThompson.h"
#include "regexp/unbounded/UnboundedRegExpElements.h"
#include "regexp/formal/FormalRegExpElements.h"
#include <tuple>

namespace regexp {

namespace convert {

automaton::EpsilonNFA < > ToAutomatonThompson::convert(const regexp::RegExp& regexp) {
	return dispatch(regexp.getData());
}

automaton::EpsilonNFA < > ToAutomatonThompson::convert(const regexp::FormalRegExp < > & regexp) {
	//FIXME use actual algorithms that implement product alternation and iteration of re over automata and remove terrible TERRIBLE hack with dummy initial state
	int nextState = 1;
	const label::Label * headArg = nullptr;
	const label::Label * tailArg = nullptr;

	automaton::EpsilonNFA < > automaton ( label::Label ( 0 ) );
	automaton.setInputAlphabet(regexp.getAlphabet());

	regexp.getRegExp().getStructure().accept < void, ToAutomatonThompson::Formal, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);

	automaton.setInitialState ( * headArg );
	automaton.setFinalStates ( std::set < label::Label > { * tailArg } );

	automaton.removeState(label::Label(0));

	return automaton;
}

auto ToAutomatonThompsonFormalRegExp = ToAutomatonThompson::RegistratorWrapper<automaton::EpsilonNFA < >, regexp::FormalRegExp < > >(ToAutomatonThompson::convert);

automaton::EpsilonNFA < > ToAutomatonThompson::convert(const regexp::UnboundedRegExp < > & regexp) {
	//FIXME use actual algorithms that implement product alternation and iteration of re over automata and remove terrible TERRIBLE hack with dummy initial state
	int nextState = 1;
	const label::Label * headArg = nullptr;
	const label::Label * tailArg = nullptr;

	automaton::EpsilonNFA < > automaton ( label::Label ( 0 ) );
	automaton.setInputAlphabet(regexp.getAlphabet());

	regexp.getRegExp().getStructure().accept < void, ToAutomatonThompson::Unbounded, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);

	automaton.setInitialState ( * headArg );
	automaton.setFinalStates ( std::set < label::Label > { * tailArg } );

	automaton.removeState(label::Label(0));

	return automaton;
}

auto ToAutomatonThompsonUnboundedRegExp = ToAutomatonThompson::RegistratorWrapper<automaton::EpsilonNFA < >, regexp::UnboundedRegExp < > >(ToAutomatonThompson::convert);

// ----------------------------------------------------------------------------

void ToAutomatonThompson::Formal::visit(const regexp::FormalRegExpAlternation < alphabet::Symbol > & alternation, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	alternation.getLeftElement().accept < void, ToAutomatonThompson::Formal, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
	automaton.addTransition(head, string::Epsilon < >::EPSILON, *headArg);
	automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, tail);

	alternation.getRightElement().accept < void, ToAutomatonThompson::Formal, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
	automaton.addTransition(head, string::Epsilon < >::EPSILON, *headArg);
	automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, tail);

	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Formal::visit(const regexp::FormalRegExpConcatenation < alphabet::Symbol > & concatenation, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	concatenation.getLeftElement().accept < void, ToAutomatonThompson::Formal, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
	const label::Label* leftHead = headArg;
	const label::Label* leftTail = tailArg;

	concatenation.getRightElement().accept < void, ToAutomatonThompson::Formal, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
	automaton.addTransition(*leftTail, string::Epsilon < >::EPSILON, *headArg);

	headArg = &(*automaton.getStates().find(*leftHead));
	// tailArg = tailArg;
}

void ToAutomatonThompson::Formal::visit(const regexp::FormalRegExpIteration < alphabet::Symbol > & iteration, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	iteration.getElement().accept < void, ToAutomatonThompson::Formal, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
	automaton.addTransition(head, string::Epsilon < >::EPSILON, *headArg);
	automaton.addTransition(head, string::Epsilon < >::EPSILON, tail);
	automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, tail);
	automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, *headArg);

	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Formal::visit(const regexp::FormalRegExpSymbol < alphabet::Symbol > & symbol, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	automaton.addTransition(head, symbol.getSymbol(), tail);
	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Formal::visit(const regexp::FormalRegExpEpsilon < alphabet::Symbol > &, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	automaton.addTransition(head, string::Epsilon < >::EPSILON, tail);
	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Formal::visit(const regexp::FormalRegExpEmpty < alphabet::Symbol > &, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

// ----------------------------------------------------------------------------

void ToAutomatonThompson::Unbounded::visit(const regexp::UnboundedRegExpAlternation < alphabet::Symbol > & alternation, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	for(const auto& element : alternation.getElements()) {
		element->accept < void, ToAutomatonThompson::Unbounded, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
		automaton.addTransition(head, string::Epsilon < >::EPSILON, *headArg);
		automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, tail);
	}

	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Unbounded::visit(const regexp::UnboundedRegExpConcatenation < alphabet::Symbol > & concatenation, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	std::vector<std::pair<const label::Label*, const label::Label*>> tails;
	for(const auto& element : concatenation.getElements()) {
		element->accept < void, ToAutomatonThompson::Unbounded, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
		tails.push_back(std::make_pair(headArg, tailArg));
	}

	for(size_t i = 1; i < tails.size(); i++)
		automaton.addTransition(*tails[i-1].second, string::Epsilon < >::EPSILON, *tails[i].first);

	headArg = tails[0].first;
	tailArg = tails[tails.size()-1].second;
}

void ToAutomatonThompson::Unbounded::visit(const regexp::UnboundedRegExpIteration < alphabet::Symbol > & iteration, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	iteration.getElement().accept < void, ToAutomatonThompson::Unbounded, automaton::EpsilonNFA < > &, int &, const label::Label * &, const label::Label * & >(automaton, nextState, headArg, tailArg);
	automaton.addTransition(head, string::Epsilon < >::EPSILON, *headArg);
	automaton.addTransition(head, string::Epsilon < >::EPSILON, tail);
	automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, tail);
	automaton.addTransition(*tailArg, string::Epsilon < >::EPSILON, *headArg);

	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Unbounded::visit(const regexp::UnboundedRegExpSymbol < alphabet::Symbol > & symbol, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	automaton.addTransition(head, symbol.getSymbol(), tail);
	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Unbounded::visit(const regexp::UnboundedRegExpEpsilon < alphabet::Symbol > &, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	automaton.addTransition(head, string::Epsilon < >::EPSILON, tail);
	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

void ToAutomatonThompson::Unbounded::visit(const regexp::UnboundedRegExpEmpty < alphabet::Symbol > &, automaton::EpsilonNFA < > & automaton , int & nextState, const label::Label * & headArg, const label::Label * & tailArg) {
	label::Label head = label::Label(nextState++);
	label::Label tail = label::Label(nextState++);
	automaton.addState(head);
	automaton.addState(tail);

	headArg = &(*automaton.getStates().find(head));
	tailArg = &(*automaton.getStates().find(tail));
}

} /* namespace convert */

} /* namespace regexp */
