/*
 * ExactFactorAutomaton.cpp
 *
 *  Created on: 7. 4. 2015
 *      Author: Jan Travnicek
 */

#include "ExactFactorAutomaton.h"
#include <string/LinearString.h>
#include <string/Epsilon.h>

#include <deque>

namespace stringology {

namespace exact {

automaton::Automaton ExactFactorAutomaton::construct(const string::String& text) {
	return dispatch(text.getData());
}

automaton::EpsilonNFA < > ExactFactorAutomaton::construct(const string::LinearString < >& text) {
	automaton::EpsilonNFA < > res(label::Label(0));
	res.addFinalState(label::Label(0));
	res.setInputAlphabet(text.getAlphabet());
	int i = 1;
	for(const alphabet::Symbol& symbol : text.getContent()) {
		res.addState(label::Label(i));
		res.addFinalState(label::Label(i));
		res.addTransition(label::Label(i-1), symbol, label::Label(i));
		res.addTransition(label::Label(0), label::Label(i));
		i++;
	}
	return res;
}

auto ExactFactorAutomatonLinearString = ExactFactorAutomaton::RegistratorWrapper<automaton::EpsilonNFA < >, string::LinearString < >>(ExactFactorAutomaton::construct);

} /* namespace exact */

} /* namespace stringology */
