#include <list>
#include "trimTest.h"

#include "automaton/simplify/Trim.h"
#include "grammar/simplify/Trim.h"

#include "automaton/FSM/DFA.h"
#include "grammar/Regular/RightRG.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( trimTest, "automaton" );
CPPUNIT_TEST_SUITE_REGISTRATION( trimTest );

void trimTest::setUp() {
}

void trimTest::tearDown() {
}

void trimTest::testTrimAutomaton() {
  automaton::DFA < > automaton(label::Label(1));

  automaton.addState(label::Label(1));
  automaton.addState(label::Label(2));
  automaton.addState(label::Label(3));
  automaton.addInputSymbol(alphabet::Symbol("a"));
  automaton.addInputSymbol(alphabet::Symbol("b"));
  
  automaton.addTransition(label::Label(1), alphabet::Symbol("a"), label::Label(2));
  automaton.addTransition(label::Label(2), alphabet::Symbol("b"), label::Label(1));
  automaton.addTransition(label::Label(3), alphabet::Symbol("b"), label::Label(1));

  automaton.addFinalState(label::Label(1));
  
  automaton::DFA<> trimed = automaton::simplify::Trim::trim(automaton);

  CPPUNIT_ASSERT(trimed.getStates().size() == 2);
}

void trimTest::testTrimGrammar() {
  grammar::RightRG < > rrGrammar(alphabet::Symbol(1));

  rrGrammar.addNonterminalSymbol(alphabet::Symbol(1));
  rrGrammar.addNonterminalSymbol(alphabet::Symbol(2));
  rrGrammar.addNonterminalSymbol(alphabet::Symbol(3));
  rrGrammar.addNonterminalSymbol(alphabet::Symbol(4));
  rrGrammar.addNonterminalSymbol(alphabet::Symbol(5));
  rrGrammar.addNonterminalSymbol(alphabet::Symbol(6));
  rrGrammar.addTerminalSymbol(alphabet::Symbol("a"));
  rrGrammar.addTerminalSymbol(alphabet::Symbol("b"));
  
  rrGrammar.addRule(alphabet::Symbol(1), std::make_pair(alphabet::Symbol("a"), alphabet::Symbol(2)));
  rrGrammar.addRule(alphabet::Symbol(2), std::make_pair(alphabet::Symbol("b"), alphabet::Symbol(3)));
  rrGrammar.addRule(alphabet::Symbol(3), alphabet::Symbol("a"));

  rrGrammar.addRule(alphabet::Symbol(4), std::make_pair(alphabet::Symbol("b"), alphabet::Symbol(5)));
  rrGrammar.addRule(alphabet::Symbol(5), alphabet::Symbol("a"));
  rrGrammar.addRule(alphabet::Symbol(5), std::make_pair(alphabet::Symbol("b"), alphabet::Symbol(2)));
  rrGrammar.addRule(alphabet::Symbol(6), std::make_pair(alphabet::Symbol("b"), alphabet::Symbol(6)));

  grammar::RightRG < > trimed = grammar::simplify::Trim::trim(rrGrammar);

  CPPUNIT_ASSERT(trimed.getNonterminalAlphabet().size() == 3);
}
