#include "LeftFactorize.h"

#include "grammar/ContextFree/CFG.h"
#include "grammar/parsing/LeftFactorize.h"

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION ( LeftFactorize, "grammar" );
CPPUNIT_TEST_SUITE_REGISTRATION ( LeftFactorize );

void LeftFactorize::setUp ( ) {
}

void LeftFactorize::tearDown ( ) {
}

void LeftFactorize::testLeftFactorize ( ) {
	alphabet::Symbol S = alphabet::Symbol ( "S" );
	alphabet::Symbol Sp = alphabet::Symbol ( "S'" );
	alphabet::Symbol B = alphabet::Symbol ( "B" );
	alphabet::Symbol Bp = alphabet::Symbol ( "B'" );
	alphabet::Symbol C = alphabet::Symbol ( "C" );
	alphabet::Symbol Cp = alphabet::Symbol ( "C'" );

	alphabet::Symbol a = alphabet::Symbol ( 'a' );
	alphabet::Symbol b = alphabet::Symbol ( 'b' );
	alphabet::Symbol c = alphabet::Symbol ( 'c' );

	grammar::CFG < > grammar ( S );

	grammar.setTerminalAlphabet ( { a, b, c } );
	grammar.setNonterminalAlphabet ( { S, B, C } );
	grammar.setInitialSymbol ( S );

	grammar.addRule ( S, std::vector < alphabet::Symbol > { a, B } );
	grammar.addRule ( S, std::vector < alphabet::Symbol > { a, C } );
	grammar.addRule ( B, std::vector < alphabet::Symbol > { b, B } );
	grammar.addRule ( B, std::vector < alphabet::Symbol > { b } );
	grammar.addRule ( C, std::vector < alphabet::Symbol > { c, C } );
	grammar.addRule ( C, std::vector < alphabet::Symbol > { c } );

	grammar::CFG < > res = grammar;
	grammar::parsing::LeftFactorize::leftFactorize ( res, a, S);
	grammar::parsing::LeftFactorize::leftFactorize ( res, b, B);
	grammar::parsing::LeftFactorize::leftFactorize ( res, c, C);

	grammar::CFG < > comp ( S );

	comp.setTerminalAlphabet ( { a, b, c } );
	comp.setNonterminalAlphabet ( { S, Sp, B, Bp, C, Cp } );
	comp.setInitialSymbol ( S );

	comp.addRule ( S, std::vector < alphabet::Symbol > { a, Sp } );
	comp.addRule ( Sp, std::vector < alphabet::Symbol > { B } );
	comp.addRule ( Sp, std::vector < alphabet::Symbol > { C } );
	comp.addRule ( B, std::vector < alphabet::Symbol > { b, Bp } );
	comp.addRule ( Bp, std::vector < alphabet::Symbol > { B } );
	comp.addRule ( Bp, std::vector < alphabet::Symbol > { } );
	comp.addRule ( C, std::vector < alphabet::Symbol > { c, Cp } );
	comp.addRule ( Cp, std::vector < alphabet::Symbol > { C } );
	comp.addRule ( Cp, std::vector < alphabet::Symbol > { } );

//	std::cout << res << std::endl << comp << std::endl;

	CPPUNIT_ASSERT ( res == comp );
}

void LeftFactorize::testLeftFactorize2 ( ) {
	alphabet::Symbol S = alphabet::Symbol ( "S" );
	alphabet::Symbol Sp = alphabet::Symbol ( "S'" );
	alphabet::Symbol B = alphabet::Symbol ( "B" );
	alphabet::Symbol Bp = alphabet::Symbol ( "B'" );
	alphabet::Symbol C = alphabet::Symbol ( "C" );
	alphabet::Symbol Cp = alphabet::Symbol ( "C'" );
	alphabet::Symbol D = alphabet::Symbol ( "D" );

	alphabet::Symbol a = alphabet::Symbol ( 'a' );
	alphabet::Symbol b = alphabet::Symbol ( 'b' );
	alphabet::Symbol c = alphabet::Symbol ( 'c' );
	alphabet::Symbol d = alphabet::Symbol ( 'd' );

	grammar::CFG < > grammar ( S );

	grammar.setTerminalAlphabet ( { a, b, c, d } );
	grammar.setNonterminalAlphabet ( { S, B, C, D } );
	grammar.setInitialSymbol ( S );

	grammar.addRule ( S, std::vector < alphabet::Symbol > { a, B } );
	grammar.addRule ( S, std::vector < alphabet::Symbol > { a, C } );
	grammar.addRule ( S, std::vector < alphabet::Symbol > { D } );
	grammar.addRule ( B, std::vector < alphabet::Symbol > { b, B } );
	grammar.addRule ( B, std::vector < alphabet::Symbol > { b } );
	grammar.addRule ( C, std::vector < alphabet::Symbol > { c, C } );
	grammar.addRule ( C, std::vector < alphabet::Symbol > { c } );
	grammar.addRule ( C, std::vector < alphabet::Symbol > { b } );
	grammar.addRule ( D, std::vector < alphabet::Symbol > { d } );

	grammar::CFG < > res = grammar;
	grammar::parsing::LeftFactorize::leftFactorize ( res, a, S);
	grammar::parsing::LeftFactorize::leftFactorize ( res, b, B);
	grammar::parsing::LeftFactorize::leftFactorize ( res, c, C);

	grammar::CFG < > comp ( S );

	comp.setTerminalAlphabet ( { a, b, c, d } );
	comp.setNonterminalAlphabet ( { S, Sp, B, Bp, C, Cp, D} );
	comp.setInitialSymbol ( S );

	comp.addRule ( S, std::vector < alphabet::Symbol > { a, Sp } );
	comp.addRule ( S, std::vector < alphabet::Symbol > { D } );
	comp.addRule ( Sp, std::vector < alphabet::Symbol > { B } );
	comp.addRule ( Sp, std::vector < alphabet::Symbol > { C } );
	comp.addRule ( B, std::vector < alphabet::Symbol > { b, Bp } );
	comp.addRule ( Bp, std::vector < alphabet::Symbol > { B } );
	comp.addRule ( Bp, std::vector < alphabet::Symbol > { } );
	comp.addRule ( C, std::vector < alphabet::Symbol > { c, Cp } );
	comp.addRule ( C, std::vector < alphabet::Symbol > { b } );
	comp.addRule ( Cp, std::vector < alphabet::Symbol > { C } );
	comp.addRule ( Cp, std::vector < alphabet::Symbol > { } );
	comp.addRule ( D, std::vector < alphabet::Symbol > { d } );

//	std::cout << res << std::endl << comp << std::endl;

	CPPUNIT_ASSERT ( res == comp );
}

