/*
 * UsefullStates.cpp
 *
 *  Created on: 23. 3. 2014
 *	  Author: Tomas Pecka
 */

#include "UsefullStates.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <queue>

namespace automaton {

namespace properties {

namespace efficient {

std::set<label::Label> UsefullStates::usefullStates(const Automaton& automaton) {
	return dispatch(automaton.getData());
}

template<class T>
std::set<label::Label> UsefullStates::usefullStates( const T & fsm ) {
	std::map<label::Label, std::set<label::Label>> reversedTransitions;
	for(const auto& transition : fsm.getTransitions())
		for(const label::Label& to : transition.second)
			reversedTransitions[to].insert(transition.first.first);

	std::deque<label::Label> queue ( fsm.getFinalStates( ).begin(), fsm.getFinalStates().end() );
	std::set<label::Label> visited = fsm.getFinalStates( );

	while( !queue.empty() ) {
		const std::set<label::Label>& to = reversedTransitions[queue.front()];
		queue.pop_front();

		for(const label::Label& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<label::Label&>(process)));
			}
	}

	return visited;
}

auto UsefullStatesEpsilonNFA = UsefullStates::RegistratorWrapper<std::set<label::Label>, automaton::EpsilonNFA < > >(UsefullStates::usefullStates);
auto UsefullStatesNFA = UsefullStates::RegistratorWrapper<std::set<label::Label>, automaton::NFA < > >(UsefullStates::usefullStates);
auto UsefullStatesCompactNFA = UsefullStates::RegistratorWrapper<std::set<label::Label>, automaton::CompactNFA < > >(UsefullStates::usefullStates);
auto UsefullStatesExtendedNFA = UsefullStates::RegistratorWrapper<std::set<label::Label>, automaton::ExtendedNFA < > >(UsefullStates::usefullStates);
auto UsefullStatesMultiInitialStateNFA = UsefullStates::RegistratorWrapper<std::set<label::Label>, automaton::MultiInitialStateNFA < >>(UsefullStates::usefullStates);

template<>
std::set<label::Label> UsefullStates::usefullStates( const automaton::DFA < > & fsm ) {
	std::map<label::Label, std::set<label::Label>> reversedTransitions;
	for(const auto& transition : fsm.getTransitions())
		reversedTransitions[transition.second].insert(transition.first.first);

	std::deque<label::Label> queue ( fsm.getFinalStates( ).begin(), fsm.getFinalStates().end() );
	std::set<label::Label> visited = fsm.getFinalStates( );

	while( !queue.empty() ) {
		const std::set<label::Label>& to = reversedTransitions[queue.front()];
		queue.pop_front();

		for(const label::Label& process : to)
			if(visited.insert(process).second) {
				queue.push_back(std::move(const_cast<label::Label&>(process)));
			}
	}

	return visited;
}

auto UsefullStatesDFA = UsefullStates::RegistratorWrapper<std::set<label::Label>, automaton::DFA<>>(UsefullStates::usefullStates);

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

